import { useRef, useState, useEffect, Fragment, createElement } from "react";
import * as ReactDom from "react-dom";
const slotToReactComponent = (renderSlot) => (props) => {
    const ref = useRef(null);
    useEffect(() => {
        renderSlot(ref.current, { key: props._key, ...props });
    });
    return createElement(props._tag || "div", { ref, key: props._key, ...props._props }, null);
};
function toReactProps(props) {
    const reactProps = { ...props.attrs, ...props.events, ...props._ };
    for (const key in props.slots) {
        const slotRenderFn = props.slots[key];
        if (!slotRenderFn)
            continue;
        reactProps[key] = slotToReactComponent(slotRenderFn);
        reactProps[key]._render = slotRenderFn;
    }
    return reactProps;
}
let uid = 0;
let forceUpdate;
export const rootElement = document.createElement("div");
const protals = new Map();
function HFCReactRoot() {
    const [, update] = useState(0);
    forceUpdate = () => update((n) => n + 1);
    return createElement(Fragment, null, Array.from(protals).map(([key, { container, node }]) => ReactDom.createPortal(node, container, key)));
}
const createRoot = ReactDom.createRoot;
if (createRoot) {
    // hijack createRoot error log
    const log = console.error;
    console.error = () => { };
    const root = createRoot(rootElement);
    console.error = log;
    root.render(createElement(HFCReactRoot));
}
else {
    ReactDom.render(createElement(HFCReactRoot), rootElement);
}
export function toHFC(Comp, opts) {
    const HFC = (container, initProps) => {
        if (opts.connected)
            opts.connected(container);
        const key = "k" + uid++;
        let changeReactProps;
        function Wrapper() {
            const [props, setProps] = useState(toReactProps(initProps));
            changeReactProps = (hfcProps) => setProps(toReactProps(hfcProps));
            return createElement(Comp, props);
        }
        protals.set(key, {
            container,
            node: createElement(Wrapper),
        });
        forceUpdate();
        return {
            changed(props) {
                changeReactProps(props);
            },
            disconnected() {
                if (opts.disconnected)
                    opts.disconnected();
                protals.delete(key);
                forceUpdate();
            },
        };
    };
    HFC.tag = opts.tag;
    HFC.hfc = opts.hfc;
    HFC.ver = opts.ver;
    HFC.names = opts.names;
    return HFC;
}
export function toHFCReact(Comp, opts) {
    const slotKeys = new Set(opts.names[2]);
    return function HFCReact(props) {
        const container = useRef(null);
        const _props = {};
        for (const key in props) {
            const value = props[key];
            // slot key
            if (slotKeys.has(key)) {
                _props[key] = toHfcSlot(value);
            }
            else {
                _props[key] = value;
            }
        }
        useEffect(() => {
            if (opts.connected)
                opts.connected(container.current);
            return () => {
                if (opts.disconnected)
                    opts.disconnected();
            };
        }, []);
        return createElement(opts.tag, { ref: container }, createElement(Comp, _props));
    };
}
function toHfcSlot(Comp) {
    return (props) => {
        return createElement(props._tag || "div", { key: props._key, ...props._props }, createElement(Comp, props));
    };
}
